# Query is used to fetch the fields of a given type to check server compatibility
query TypeFields (
    $typeName: String!
) {
    TypeInfo: __type(name: $typeName) {
        fields {
            name
        }
    }
}
